/************************************************/
/* File:	barbershop.c 						*/
/* Author:	Sean Handley, 320097@swan.ac.uk 	*/
/* Date:	2006.03.18 							*/
/************************************************/

#define _XOPEN_SOURCE 1

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "shop_subs.h"
#include "barbershop.h"

#define BARBER_KEY (key_t)75
#define CUSTOMER_KEY (key_t)125
#define WC_KEY (key_t) 150
#define SCREEN_KEY (key_t) 175

#define DEBUG 0

#define MAX_CUST_LIMIT 10
#define MAX_CHAIR_LIMIT 10
#define DEBUG_LOG "debug.log"

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/*already defined*/
#else
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
	struct seminfo *__buf;
};
#endif

/* semaphore ids */
int barb_id, cust_id, wc_id, screen_id;
struct sembuf barbop_p, barbop_v;
struct sembuf custop_p, custop_v;
struct sembuf wcop_p, wcop_v;
struct sembuf screenop_p, screenop_v;
union semun barb_arg, cust_arg, wc_arg, screen_arg;
int barb_val, cust_val, wc_val, screen_val;

int wait_count;

/* utility error msg function */
void errormsg(char *msg) {
	fprintf(stderr,"%s\n",msg);
	exit(0);
}

/* utility debug msg function */
void debugmsg(char *msg) {
	fprintf(stdout,"DEBUG: %s\n",msg);
}

void debuglog(char *msg) {
	FILE *output = fopen(DEBUG_LOG,"a");
	fprintf(output,"%s\n",msg);
	fclose(output);
}

void debuglogparam(char *msg, int param) {
	FILE *output = fopen(DEBUG_LOG,"a");
	fprintf(output,"%s : %d\n",msg,param);
	fclose(output);
}

/* utility input validation function */
/* keeps numbers sane and input secure */
int validate(char *argv[]) {
	int success = 1;

		if(atoi(argv[1]) > MAX_CUST_LIMIT) {
			success = 0;
			errormsg("Too many customers!");
		}

		if(atoi(argv[2]) > MAX_CHAIR_LIMIT) {
			success = 0;
			errormsg("Too many chairs!");
		}

		if((atoi(argv[3]) > 1) || (atoi(argv[3]) < 0)) {
			success = 0;
			errormsg("Mutual exclusion flag must be 0 or 1.");
		}
	
	return success;
}

void initialise_semaphores() {
	/* we need to set up 3 separate semaphores */
	/* barber, customer and wait_count */

	/* to catch Ctrl+C */
	signal(SIGINT, cleanup_semaphore);

	/* zero the wait count */
	wait_count = 0;

#if DEBUG
	debuglog("Initialising barber semaphore...");
#endif

	barb_id = semget(BARBER_KEY, 1, 0666|IPC_CREAT);
	if(barb_id == -1) {
		errormsg("Failed to get semaphore.");
		exit(0);
	} else {
#if DEBUG
		debuglog("Barber semaphore created.");
#endif
	}

	barb_arg.val = 0;
	semctl(barb_id, 0, SETVAL, barb_arg);

#if DEBUG
	debuglog("Initialising customer semaphore...");
#endif

	cust_id = semget(CUSTOMER_KEY, 1, 0666|IPC_CREAT);
	if(cust_id == -1) {
	    errormsg("Failed to get semaphore.");
		exit(0);
    } else {
#if DEBUG
	    debuglog("Customer semaphore created.");
#endif
    }

    cust_arg.val = 0;
	semctl(cust_id, 0, SETVAL, cust_arg);


#if DEBUG
    debuglog("Initialising wait count semaphore...");
#endif

    wc_id = semget(WC_KEY, 1, 0666|IPC_CREAT);
    if(wc_id == -1) {
        errormsg("Failed to get semaphore.");
        exit(0);
    } else {
#if DEBUG
        debuglog("Wait count semaphore created.");
#endif
    }

    wc_arg.val = 1;
    semctl(wc_id, 0, SETVAL, wc_arg);
	
#if DEBUG
	debuglog("Initialising screen semaphore...");
#endif

	screen_id = semget(SCREEN_KEY, 1, 0666|IPC_CREAT);
	if(screen_id == -1) {
		errormsg("Failed to get semaphore.");
		exit(0);
	} else {
#if DEBUG
		debuglog("Screen semaphore.");
#endif
	}

	screen_arg.val = 0;
	semctl(screen_id, 0, SETVAL, screen_arg);
		
	barbop_p.sem_num = 0;
	barbop_p.sem_op = -1;
	barbop_p.sem_flg = SEM_UNDO;

	barbop_v.sem_num = 0;
	barbop_v.sem_op = 1;
	barbop_v.sem_flg = SEM_UNDO;
	
	custop_p.sem_num = 0;
	custop_p.sem_op = -1;
	custop_p.sem_flg = SEM_UNDO;
	
	custop_v.sem_num = 0;
	custop_v.sem_op = 1;
	custop_v.sem_flg = SEM_UNDO;

    wcop_p.sem_num = 0;
    wcop_p.sem_op = -1;
	wcop_p.sem_flg = SEM_UNDO;

    wcop_v.sem_num = 0;
    wcop_v.sem_op = 1;
    wcop_v.sem_flg = SEM_UNDO;

	screenop_p.sem_num = 0;
	screenop_p.sem_op = -1;
	screenop_p.sem_flg = SEM_UNDO;

	screenop_v.sem_num = 0;
	screenop_v.sem_op = 1;
	screenop_v.sem_flg = SEM_UNDO;

	/*create handles*/
	barb_id = semget(BARBER_KEY, 1, 0666);
	cust_id = semget(CUSTOMER_KEY, 1, 0666);
	wc_id = semget(WC_KEY, 1, 0666);
	screen_id = semget(SCREEN_KEY, 1, 0666);
	
#if DEBUG
    debuglogparam("INIT: cust sem val",semctl(cust_id, 0, GETVAL));
    debuglogparam("INIT: barb sem val",semctl(barb_id, 0, GETVAL));
    debuglogparam("INIT: wc sem val",semctl(wc_id, 0, GETVAL));
#endif
													
	
}

void cleanup_semaphore() {
	semctl(barb_id, 0, IPC_RMID, 0);
	semctl(cust_id, 0, IPC_RMID, 0);
	semctl(wc_id, 0, IPC_RMID, 0);
}

void barber_end() {
#if DEBUG
	debuglog("Barber is leaving the shop.");
#endif
	draw_message("Barber leaves");
	barber_leaves_barbershop();
	exit(0);
}

/* where the barbershop simulation gets executed */
void run(int customers, int chairs, int mut_exc) {

	int i;
	pid_t cust_pid[MAX_CUST_LIMIT];
	pid_t barber_pid;

	initialise_semaphores();
	
#if DEBUG
	debuglogparam("INIT2: cust sem val",semctl(cust_id, 0, GETVAL));
	debuglogparam("INIT2: barb sem val",semctl(barb_id, 0, GETVAL));
	debuglogparam("INIT2: wc sem val",semctl(wc_id, 0, GETVAL));
#endif

	/* set up shop */
	open_barbershop(customers,chairs,mut_exc);

#if DEBUG
	debuglogparam("INIT3: cust sem val",semctl(cust_id, 0, GETVAL));
	debuglogparam("INIT3: barb sem val",semctl(barb_id, 0, GETVAL));
	debuglogparam("INIT3: wc sem val",semctl(wc_id, 0, GETVAL));
#endif

	draw_message("At time T0: open the barbershop");
	
	barber_pid = fork();
	if(barber_pid == 0) {
		/* this is the barber */
#if DEBUG
		debuglog("Barber has been forked.");
#endif
		/* set up signal response */
		signal(SIGUSR1,barber_end);
		barber_enters_barbershop();
		barber_goes_to_sleep();
		barber_actions(mut_exc);
		
	} else if(barber_pid > 0) {
		/* this is the program */
		/* fork the customers */
		for(i = 0; i < customers; i++) {
			cust_pid[i] = fork();
			if(cust_pid[i] == 0) {
				/* this is the customer */
#if DEBUG
				debuglog("Customer has been forked.");
#endif
				customer_actions(chairs, mut_exc);
			} else if(cust_pid[i] > 0) {
				continue;	
			} else {
				errormsg("Failed to fork.");
			}
		}
	} else {
		errormsg("Failed to fork.");
	}
	/* this is the program */
	/* wait for children to die */
	for(i=0; i < customers; i++) {	
		while(waitpid(cust_pid[i],NULL,WNOHANG) != -1) {};
	}
	
#if DEBUG
	debuglog("Customers gone.");
	debuglog(strerror(errno));
#endif
	
	kill(barber_pid,SIGUSR1);
	cleanup_semaphore();
	sleep(2);
	close_barbershop();
	exit(0);
}

/* what the barber does */
void barber_actions(int mut_exc) {
	if(mut_exc == MUTUAL_EXCLUSION_ON) {
		while(1) {
#if DEBUG
			debuglog("Barber sleeps.");
			debuglogparam("cust sem val",semctl(cust_id, 0, GETVAL));
			debuglogparam("barb sem val",semctl(barb_id, 0, GETVAL));
			debuglogparam("wc sem val",semctl(wc_id, 0, GETVAL));
#endif
			semop(cust_id, &custop_p, 1);

			semop(wc_id, &wcop_p, 1);
			--wait_count;
			semop(wc_id, &wcop_v, 1);

#if DEBUG
			debuglog("Barber cuts hair.");
#endif
		}
	} else {
		/* no mutual exclusion */
		barber_goes_to_sleep();
	}
}

/* what the customer does */
void customer_actions(int chairs, int mut_exc) {

	if(mut_exc == MUTUAL_EXCLUSION_ON) {
		/* semop(screen_id, &screenop_p, 1); */
		draw_message("Customer enters barbershop.");
		customer_enters_barbershop(getpid());
		/* semop(screen_id, &screenop_v, 1); */
#if DEBUG
		debuglog("Customer enters the barbershop.");
#endif
		semop(wc_id, &wcop_p, 1);
		if(wait_count < chairs) {
			++wait_count;
		/*	semop(screen_id, &screenop_p, 1);*/
			draw_message("Customer gets waiting chair");
			customer_gets_waiting_chair(getpid());
/*			semop(screen_id, &screenop_v, 1);*/
#if DEBUG
			debuglog("Customer gets waiting chair.");
#endif
			semop(cust_id, &custop_v, 1);
			semop(wc_id, &wcop_v, 1);
			semop(barb_id, &barbop_p, 1);

/*			semop(screen_id, &screenop_p, 1);*/
			draw_message("Customer gets barber chair.");
            customer_gets_barber_chair(getpid());
			barber_becomes_busy();
			draw_message("Customer's hair gets cut.");
            customer_has_haircut(getpid());
			draw_message("Barber sleeps");
			barber_goes_to_sleep();
/*			semop(screen_id, &screenop_v, 1);*/
#if DEBUG
			debuglog("Customer gets a haircut.");
#endif
		} else {
			semop(wc_id, &wcop_v, 1);
		}
/*		semop(screen_id, &screenop_p, 1);*/
		customer_leaves_barbershop(getpid());
		draw_message("Customer leaves.");
/*		semop(screen_id, &screenop_v, 1);*/
#if DEBUG
		debuglog("Customer leaves.");
#endif
		exit(0);
	} else {
		/*no mutual exclusion
 		customer_enters_barbershop(getpid());
		if(wait_count < chairs) {
			customer_gets_waiting_chair(getpid());
			customer_gets_barber_chair(getpid());
			barber_becomes_busy();
			customer_has_haircut(getpid());
		}
		customer_leaves_barbershop(getpid());
		exit(0);*/
	}
}

int main(int argc, char *argv[]) {
#if DEBUG
	/* if we're in debug mode, clear the log */
	FILE *output = fopen(DEBUG_LOG,"w");
	fprintf(output,"\n");
	fclose(output);
#endif
	/* validate argc */
	switch(argc) {
		/* 4 args needed to run */
		case 4:
#if DEBUG
			debugmsg("arg count correct");
#endif
			/* validate individual args */
			if(validate(argv)) {
#if DEBUG
				debugmsg("argv is valid");
#endif
				/* input is valid; proceed */
				run(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]));
			}
			break;
		default:
			errormsg("Incorrect usage.\nFormat of command is:\n./barbershop <int> <int> <bool>\n\nsuch that\n\n./barbershop <no_of_customers> <no_of_chairs> <mutual_exclusion>\n");
	}
	return 0;
}
