/* ==================================================================== */
/* File:	shop_subs.c						*/
/* Author:	Dr. M. Chen, University of Wales Swansea		*/
/* Date:	First created on 31/1/95				*/
/* Description:	This file consists of a set of subroutines, built on	*/
/*		the top of the curses library, may be used for the	*/
/*		operating system coursework on the sleeping barber	*/
/*		problem. There are five groups of routines as follows:	*/
/*									*/
/*		Group A includes subroutines which are used to activate	*/
/*		and deactivate the display environment for the problem.	*/
/*									*/
/*		Group B includes subroutines which are used to display	*/
/*		the status of the barber process.			*/
/*									*/
/*		Group C includes subroutines which are used to display	*/
/*		the status of a customer process.			*/
/*									*/
/*		Group D includes low level drawing procedures which	*/
/*		have been designed to allow several processes to write	*/
/*		concurrently on the same screen using curses routines.	*/
/*									*/
/*		Group E includes semaphore operations used mainly by	*/
/*		Group D subroutines to ensure the consistency.		*/
/* ==================================================================== */

/* Under Linux, you need to include this 'define, otherwise you get a
 *  warning when you include <sys/ipc.h>
 */
#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <unistd.h>
#include <curses.h>

#include "shop_subs.h"

union semun {
    int val;
    struct semid_ds *buf;
    unsigned short *array;
    struct seminfo *__buf;
    };

/* -------------------------------------------------------------------- */
/* Constants that are used to control the speed of display.		*/
/* -------------------------------------------------------------------- */
#define DIST_TO_WAITING_CHAIR	10	/* value range [1..10] in m */
#define DIST_TO_BARBER_CHAIR	10	/* value range [1..10] in m */
#define HAIR_LENGTH		10	/* value range [1..10] in cm */
#define WALKING_TIME_PER_M	130000	/* positive integer */
#define CUTTING_TIME_PER_CM	110000	/* positive integer */

/* -------------------------------------------------------------------- */
/* Constants that define symbols representing different statuses.	*/
/* -------------------------------------------------------------------- */
#define SPACE_SYMBOL		' '
#define ENTERING_SYMBOL		'{'
#define LEAVING_SYMBOL		'}'
#define WAITING_CHAIR_SYMBOL	'W'
#define BARBER_CHAIR_SYMBOL	'B'
#define WALKING_SYMBOL		'>'
#define HAIRCUTTING_SYMBOL	'~'
#define PAYMENT_SYMBOL		'$'

/* -------------------------------------------------------------------- */
/* Constants that define the maximum numbers of barbers, customers and	*/
/*		waiting chairs. The number should be set according to	*/
/*		the number of the available text lines on the screen.	*/
/* -------------------------------------------------------------------- */
#define MAX_BARBERS		1	/* not checked nor used */
#define MAX_CUSTOMERS		20	/* require 15+9 lines on screen */
#define MAX_WAITING_CHAIRS	20	/* checked but not used */

/* -------------------------------------------------------------------- */
/* Constants and a Macro that define the y position (the line no.) of 	*/
/*		each child process and a general message line.		*/
/* -------------------------------------------------------------------- */
#define BARBER_LINE_OFFSET	8
#define CUSTOMER_LINE_OFFSET	(BARBER_LINE_OFFSET+MAX_BARBERS)
#define CustomerLine(x)		(x + CUSTOMER_LINE_OFFSET)
#define MESSAGE_LINE_OFFSET	(CUSTOMER_LINE_OFFSET+MAX_CUSTOMERS)

/* -------------------------------------------------------------------- */
/* Constants that define x positions for status symbols.		*/
/* -------------------------------------------------------------------- */
#define ENTERING_OFFSET		19
#define BARBER_STATE_OFFSET	43
#define WAITING_CHAIR_OFFSET	32
#define BARBER_CHAIR_OFFSET	45
#define PAYMENT_OFFSET		58
#define LEAVING_OFFSET		60

/* -------------------------------------------------------------------- */
/* Constant strings that describe various status symbols.		*/
/* -------------------------------------------------------------------- */
#define SYMBOLS_DESCRIPTION1	"  `{' Enter, `W' Waiting Chair, `~' Haircut,"
#define SYMBOLS_DESCRIPTION2	"`}' Leave, `B' Barber Chair, `>' Walk, `$' Pay"

/* -------------------------------------------------------------------- */
/* Miscellaneous constants.						*/
/* -------------------------------------------------------------------- */
#define TMP_STR_LEN		256	/* the size of a temporary string */

/* -------------------------------------------------------------------- */
/* Local semaphore operations.						*/
/* -------------------------------------------------------------------- */

static int create_semaphore();
static void destroy_semaphore();
static void semaphore_p();
static void semaphore_v();

/* ==================================================================== */
/* Group A: DISPLAY ENVIRONMENT						*/
/* ==================================================================== */

/* ---------------------------------------------------- */
/* Function:	open_barbershop(customers, wchairs).	*/
/* Description: activate the display environment for a	*/
/*		barbershop by creating a semaphore,	*/
/*		initialising the curses library and 	*/
/*		displaying headings.			*/
/* Parameters:  customers -- the number of customers.	*/
/*		wchairs -- the number of waiting chairs.*/
/*		flag -- indicating whether or not	*/
/*		mutual exclusion is enforced in user's	*/
/*		barbershop program. (0 = off, 1 = on)	*/
/* Return:	0 -- OK, -1 -- Error.			*/
/* ---------------------------------------------------- */

int open_barbershop(int customers, int wchairs, int flag)
{
	if (customers > MAX_CUSTOMERS || wchairs > MAX_WAITING_CHAIRS)
		return(-1);
	else if (customers < 1 || wchairs < 1)
		return(-1);

	/* create a local semaphore for Group D routines. */
	if (create_semaphore() < 0)
		return(-1);

	/* initialise the curses library. */
	initscr();

	/* display headings. */
	draw_headings(customers, wchairs, flag);

	return(0);
}

/* ---------------------------------------------------- */
/* Function:	close_barbershop().			*/
/* Description: deactivate the display environment.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void close_barbershop()
{
	/* move the cursor the bottom of the display and finish up */
	/* the curses routines. */
	move(MESSAGE_LINE_OFFSET+1, 0);
	refresh();
	endwin();

	/* clean up the semaphore. */
	destroy_semaphore();
}

/* ==================================================================== */
/* Group B: BARBER INFORMATION						*/
/* ==================================================================== */

/* ---------------------------------------------------- */
/* Function:	barber_enters_barbershop().		*/
/* Description: the barber enters the barbershop.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void barber_enters_barbershop()
{
	char	str[TMP_STR_LEN];

	/* Seed the random number generator. */
	srand(getpid());

	/* Display the pid of the barber process. */
	sprintf(str, "Barber   (%d): ", getpid());
	draw_string(0, BARBER_LINE_OFFSET, str);

	/* The barber enters the barbershop. */
	draw_char(ENTERING_OFFSET, BARBER_LINE_OFFSET, ENTERING_SYMBOL);
}

/* ---------------------------------------------------- */
/* Function:	barber_leaves_barbershop().		*/
/* Description: the barber leaves the barbershop.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void barber_leaves_barbershop()
{
	/* The barber leaves the barbershop. */
	draw_char(LEAVING_OFFSET, BARBER_LINE_OFFSET, LEAVING_SYMBOL);
}

/* ---------------------------------------------------- */
/* Function:	barber_becomes_busy().			*/
/* Description: the barber starts cutting hair.		*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void barber_becomes_busy()
{
	/* The barber starts to cut a customer's hair. */
	draw_string(BARBER_STATE_OFFSET, BARBER_LINE_OFFSET, "BUSY!");
}

/* ---------------------------------------------------- */
/* Function:	barber_goes_to_sleep().			*/
/* Description: the barber goes to sleep.		*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void barber_goes_to_sleep()
{
	/* The barber goes to sleep as no customer to be served. */
	draw_string(BARBER_STATE_OFFSET, BARBER_LINE_OFFSET, "SLEEP");
}

/* ==================================================================== */
/* Group C: CUSTOMER INFORMATION					*/
/* ==================================================================== */

/* -------------------------------------------------------------------- */
/* A constant for the frequent use of a Macro CustomerLine(customer).	*/
/* -------------------------------------------------------------------- */
#define LINE	CustomerLine(customer)

/* ---------------------------------------------------- */
/* Function:	customer_enters_barbershop(customer).	*/
/* Description: a customer enters the barbershop.	*/
/* Parameters:  customer -- the customer no., its value	*/
/*		range is [0, MAX_CUSTOMERS-1].		*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

#define MAX_DELAY	11

void customer_enters_barbershop(int customer)
{
	char	str[TMP_STR_LEN];
	int	delay;

	/* Display the pid of the customer process. */
	sprintf(str, "Customer (%d): ", customer);
	draw_string(0, LINE, str);

	delay = rand() % MAX_DELAY;
	delay = delay * delay * delay / 20;
	usleep(delay);

	/* The customer enters the barbershop. */
	draw_char(ENTERING_OFFSET, LINE, ENTERING_SYMBOL);
}

/* ---------------------------------------------------- */
/* Function:	customer_leaves_barbershop(customer).	*/
/* Description: a customer leaves the barbershop.	*/
/* Parameters:  customer -- the customer no.		*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void customer_leaves_barbershop(int customer)
{
	/* The customer leaves the barbershop. */
	draw_char(LEAVING_OFFSET, LINE, LEAVING_SYMBOL);
}

/* ---------------------------------------------------- */
/* Function:	customer_gets_waiting_chair(customer).	*/
/* Description: a customer walks to and sits on a wait-	*/
/*		ing chair in the barbershop.		*/
/* Parameters:  customer -- the customer no.		*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void customer_gets_waiting_chair(int customer)
{
	int	i;

	/* The customer walks towards a waiting chair. */
	for (i=0; i<DIST_TO_WAITING_CHAIR; i++) {
		usleep(WALKING_TIME_PER_M);
		draw_char(ENTERING_OFFSET+2+i, LINE, WALKING_SYMBOL);
	}

	/* The customer sits on the waiting chair. */
	draw_char(WAITING_CHAIR_OFFSET, LINE, WAITING_CHAIR_SYMBOL);
	usleep(WALKING_TIME_PER_M);
}

/* ---------------------------------------------------- */
/* Function:	customer_gets_barber_chair(customer).	*/
/* Description: a customer walks to and sits on a/the	*/
/*		barber chair in the barbershop.		*/
/* Parameters:  customer -- the customer no.		*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void customer_gets_barber_chair(int customer)
{
	int	i;

	/* The customer releases his/her waiting chair. */
	draw_char(WAITING_CHAIR_OFFSET, LINE, SPACE_SYMBOL);

	/* The customer walks towards the barber chair. */
	for (i=0; i<DIST_TO_WAITING_CHAIR; i++) {
		draw_char(WAITING_CHAIR_OFFSET+2+i, LINE, WALKING_SYMBOL);
		usleep(WALKING_TIME_PER_M);
	}

	/* The customer sits on the barber chair. */
	draw_char(BARBER_CHAIR_OFFSET, LINE, BARBER_CHAIR_SYMBOL);
}

/* ---------------------------------------------------- */
/* Function:	customer_has_haircut(customer).		*/
/* Description: a customer has his/her haircut while	*/
/*		sitting on a/the barber chair.		*/
/* Parameters:  customer -- the customer no., not used.	*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void customer_has_haircut(int customer)
{
	int	i;

	/* The customer has a haircut. */
	for (i=0; i<HAIR_LENGTH; i++) {
		draw_char(BARBER_CHAIR_OFFSET+2+i, LINE, HAIRCUTTING_SYMBOL);
		usleep(CUTTING_TIME_PER_CM);
	}

	/* The customer pays his/her bill. */
	draw_char(PAYMENT_OFFSET, LINE, PAYMENT_SYMBOL);

	/* The customer releases the barber chair. */
	draw_char(BARBER_CHAIR_OFFSET, LINE, SPACE_SYMBOL);
}

/* ==================================================================== */
/* Group D: DRAWING ROUTINES						*/
/*	Note: In order to ensure the consistency in the curses' runtime	*/
/*	data structures maintained by different processes, semaphore	*/
/*	operations are used to make sure that each process completes	*/
/*	a drawing operation without any break. A 'move(0, 0);' call is	*/
/*	used to avoid the confusion in determining the current drawing	*/
/*	or cursor position. This method may not be 100% reliable.	*/
/* ==================================================================== */

/* ---------------------------------------------------- */
/* Function:	draw_headings(customers, wchairs).	*/
/* Description: display headings. Because it is called	*/
/*		only by the parent processes before	*/
/*		creating its child processes, mutual	*/
/*		exclusion is unnecessary.		*/
/* Parameters:  customers -- the number of customers.	*/
/*		wchairs -- the number of waiting chairs.*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void draw_headings(int customers, int wchairs, int flag)
{
	int	x, y;

	for (x=0; x<=60; x++) {
		mvaddch(0, x, '#');
		mvaddch(6, x, '#');
	}
	for (y=0; y<=6; y++) {
		mvaddch(y, 0, '#');
		mvaddch(y, 60, '#');
	}

	mvaddstr(1, 10, "MY BARBERSHOP");
	if (flag == 0)
		mvprintw(1, 26, "[Without Mutual Exclusion]");
	else	mvprintw(1, 26, "   [With Mutual Exclusion]");

	if (customers == 1)
		mvprintw(2, 14, " 1 Customer,");
	else	mvprintw(2, 14, "%2d Customers,", customers);

	if (wchairs == 1)
		mvprintw(2, 29, " 1 Waiting Chair");
	else	mvprintw(2, 29, "%2d Waiting Chairs", wchairs);
	
	mvaddstr(4, 7, SYMBOLS_DESCRIPTION1);
	mvaddstr(5, 7, SYMBOLS_DESCRIPTION2);

	move(0, 0);
	refresh();
}

/* ---------------------------------------------------- */
/* Function:	draw_char(x, y, ch).			*/
/* Description: draw a character onto the screen;	*/
/*		note the order of (x, y) coordinate.	*/
/* Parameters:  x, y -- the position of the character.	*/
/*		ch -- the character.			*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void draw_char(int x, int y, char ch)
{
	semaphore_p();
	mvaddch(y, x, ch);
	move(0, 0);
	refresh();
	semaphore_v();
}

/* ---------------------------------------------------- */
/* Function:	draw_string(x, y, str).			*/
/* Description: draw a string onto the screen;		*/
/*		note the order of (x, y) coordinate.	*/
/* Parameters:  x, y -- the position of the string.	*/
/*		str -- the string.			*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void draw_string(int x, int y, char *str)
{
	semaphore_p();
	mvaddstr(y, x, str);
	move(0, 0);
	refresh();
	semaphore_v();
}

/* ---------------------------------------------------- */
/* Function:	draw_message(str).			*/
/* Description: draw a string on the message line.	*/
/* Parameters:  str -- the string.			*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

void draw_message(char *str)
{
	semaphore_p();
	move(MESSAGE_LINE_OFFSET, 0);
	deleteln();
	mvaddstr(MESSAGE_LINE_OFFSET, 0, str);
	move(0, 0);
	refresh();
	semaphore_v();
}

/* ==================================================================== */
/* Group E: SEMAPHORE OPERATIONS					*/
/* ==================================================================== */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

/* -------------------------------------------------------------------- */
/* A constant and global variables for semaphore operations.		*/
/* -------------------------------------------------------------------- */

#define SEMAPHOREKEY	(key_t)100
static int		semid;
static struct sembuf	semop_p, semop_v;

/* ---------------------------------------------------- */
/* Function:	create_semaphore().			*/
/* Description: create and initialise a semaphore.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

static int create_semaphore()
{
	union semun semarg;

	semid = semget(SEMAPHOREKEY, 1, 0666|IPC_CREAT);
	if (semid == -1)
		return(-1);

	semarg.val = 1;
	semctl(semid, 0, SETVAL, semarg);

	semop_p.sem_num = 0;
	semop_p.sem_op = -1;
	semop_p.sem_flg = SEM_UNDO;

	semop_v.sem_num = 0;
	semop_v.sem_op = 1;
	semop_v.sem_flg = SEM_UNDO;

	return(0);
}

/* ---------------------------------------------------- */
/* Function:	destroy_semaphore().			*/
/* Description: remove a semaphore.			*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

static void destroy_semaphore()
{
	semctl(semid, 0, IPC_RMID, 0);
}

/* ---------------------------------------------------- */
/* Function:	semaphore_p().				*/
/* Description: perform a semaphore P operation.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

static void semaphore_p()
{
	semop(semid, &semop_p, 1);
}

/* ---------------------------------------------------- */
/* Function:	semaphore_v().				*/
/* Description: perform a semaphore V operation.	*/
/* Parameters:  <none>.					*/
/* Return:	<none>.					*/
/* ---------------------------------------------------- */

static void semaphore_v()
{
	semop(semid, &semop_v, 1);
}

/* ==================================================================== */
/* END									*/
/* ==================================================================== */
